import logging
import time
from functools import lru_cache

import mock
from odoo.modules.registry import Registry

from odoo_commands.database_mock import cursor_mock_class, cursor

s1 = time.time()

import odoo
from odoo.api import Environments

s2 = time.time()
print('S >>', format(s2 - s1, '.3f'))

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

cr = None


class Project:
    def __init__(self):
        self._env_generator = None

    # TODO Use functools.cached_property for Python3.8+
    @property
    @lru_cache(maxsize=1)
    def env(self):
        # Keep ref on generator force close generator later
        self._env_generator = self._get_env_generator()
        return next(self._env_generator)

    def _get_env_generator(self):
        installed_modules = [
            'account',
            'analytic',
            'base',
            'product',
            'portal',
            'base_setup',
            'mail',
            'http_routing',
            'decimal_precision',
            'web',
            'bus',
            'web_tour',
            'sale',
            'sales_team',
            'web_planner',
        ]

        with mock.patch('odoo.sql_db.Cursor', cursor_mock_class(lambda self: ['base'])):
            with odoo.api.Environment.manage():
                with odoo.registry('soma').cursor() as cr:
                    print('2')
                    yield odoo.api.Environment(cr, odoo.SUPERUSER_ID, {})

    def payload(self):
        Registry.in_test_mode = lambda self: True

        print('1')
        print(self.env)
        print('3')
        print('10')


project = Project()
project.payload()



import re
import codecs
import logging
from datetime import datetime
from lxml import etree

_logger = logging.getLogger(__name__)


SKIPPED_ELEMENT_TYPES = (etree._Comment, etree._ProcessingInstruction, etree.CommentBase, etree.PIBase, etree._Entity)
SKIPPED_ELEMENTS = ('script', 'style', 'title')


def _push(callback, term, source_line):
    """ Sanity check before pushing translation terms """
    term = (term or "").strip()
    # Avoid non-char tokens like ':' '...' '.00' etc.
    if len(term) > 8 or any(x.isalpha() for x in term):
        callback(term, source_line)


def _extract_translatable_qweb_terms(element, callback):
    """ Helper method to walk an etree document representing
        a QWeb template, and call ``callback(term)`` for each
        translatable term that is found in the document.

        :param etree._Element element: root of etree document to extract terms from
        :param Callable callback: a callable in the form ``f(term, source_line)``,
                                  that will be called for each extracted term.
    """
    # not using elementTree.iterparse because we need to skip sub-trees in case
    # the ancestor element had a reason to be skipped
    for el in element:
        if isinstance(el, SKIPPED_ELEMENT_TYPES): continue
        if (el.tag.lower() not in SKIPPED_ELEMENTS
                and "t-js" not in el.attrib
                and not ("t-jquery" in el.attrib and "t-operation" not in el.attrib)
                and el.get("t-translation", '').strip() != "off"):
            _push(callback, el.text, el.sourceline)
            for att in ('title', 'alt', 'label', 'placeholder'):
                if att in el.attrib:
                    _push(callback, el.attrib[att], el.sourceline)
            _extract_translatable_qweb_terms(el, callback)
        _push(callback, el.tail, el.sourceline)


def babel_extract_qweb(fileobj, keywords, comment_tags, options):
    """Babel message extractor for qweb template files.

    :param fileobj: the file-like object the messages should be extracted from
    :param keywords: a list of keywords (i.e. function names) that should
                     be recognized as translation functions
    :param comment_tags: a list of translator tags to search for and
                         include in the results
    :param options: a dictionary of additional options (optional)
    :return: an iterator over ``(lineno, funcname, message, comments)``
             tuples
    :rtype: Iterable
    """
    result = []
    def handle_text(text, lineno):
        result.append((lineno, None, text, []))
    tree = etree.parse(fileobj)
    _extract_translatable_qweb_terms(tree.getroot(), handle_text)
    return result



def quote(s):
    """Returns quoted PO term string, with special PO characters escaped"""
    assert r"\n" not in s, "Translation terms may not include escaped newlines ('\\n'), please use only literal newlines! (in '%s')" % s
    return '"%s"' % s.replace('\\','\\\\') \
                     .replace('"','\\"') \
                     .replace('\n', '\\n"\n"')

re_escaped_char = re.compile(r"(\\.)")
re_escaped_replacements = {'n': '\n', 't': '\t',}

def _sub_replacement(match_obj):
    return re_escaped_replacements.get(match_obj.group(1)[1], match_obj.group(1)[1])

def unquote(str):
    """Returns unquoted PO term string, with special PO characters unescaped"""
    return re_escaped_char.sub(_sub_replacement, str[1:-1])


class PoFile(object):
    def __init__(self, buffer):
        # TextIOWrapper closes its underlying buffer on close *and* can't
        # handle actual file objects (on python 2)
        self.buffer = codecs.StreamReaderWriter(
            stream=buffer,
            Reader=codecs.getreader('utf-8'),
            Writer=codecs.getwriter('utf-8')
        )

    def __iter__(self):
        self.buffer.seek(0)
        self.lines = self._get_lines()
        self.lines_count = len(self.lines)

        self.first = True
        self.extra_lines= []
        return self

    def _get_lines(self):
        lines = self.buffer.readlines()
        # remove the BOM (Byte Order Mark):
        if len(lines):
            lines[0] = lines[0].lstrip(u"\ufeff")

        lines.append('') # ensure that the file ends with at least an empty line
        return lines

    def cur_line(self):
        return self.lines_count - len(self.lines)

    def next(self):
        trans_type = name = res_id = source = trad = None
        if self.extra_lines:
            trans_type, name, res_id, source, trad, comments = self.extra_lines.pop(0)
            if not res_id:
                res_id = '0'
        else:
            comments = []
            targets = []
            line = None
            fuzzy = False
            while not line:
                if 0 == len(self.lines):
                    raise StopIteration()
                line = self.lines.pop(0).strip()
            while line.startswith('#'):
                if line.startswith('#~ '):
                    break
                if line.startswith('#.'):
                    line = line[2:].strip()
                    if not line.startswith('module:'):
                        comments.append(line)
                elif line.startswith('#:'):
                    # Process the `reference` comments. Each line can specify
                    # multiple targets (e.g. model, view, code, selection,
                    # ...). For each target, we will return an additional
                    # entry.
                    for lpart in line[2:].strip().split(' '):
                        trans_info = lpart.strip().split(':',2)
                        if trans_info and len(trans_info) == 2:
                            # looks like the translation trans_type is missing, which is not
                            # unexpected because it is not a GetText standard. Default: 'code'
                            trans_info[:0] = ['code']
                        if trans_info and len(trans_info) == 3:
                            # this is a ref line holding the destination info (model, field, record)
                            targets.append(trans_info)
                elif line.startswith('#,') and (line[2:].strip() == 'fuzzy'):
                    fuzzy = True
                line = self.lines.pop(0).strip()
            if not self.lines:
                raise StopIteration()
            while not line:
                # allow empty lines between comments and msgid
                line = self.lines.pop(0).strip()
            if line.startswith('#~ '):
                while line.startswith('#~ ') or not line.strip():
                    if 0 == len(self.lines):
                        raise StopIteration()
                    line = self.lines.pop(0)
                # This has been a deprecated entry, don't return anything
                return next(self)

            if not line.startswith('msgid'):
                raise Exception("malformed file: bad line: %s" % line)
            source = unquote(line[6:])
            line = self.lines.pop(0).strip()
            if not source and self.first:
                self.first = False
                # if the source is "" and it's the first msgid, it's the special
                # msgstr with the informations about the traduction and the
                # traductor; we skip it
                self.extra_lines = []
                while line:
                    line = self.lines.pop(0).strip()
                return next(self)

            while not line.startswith('msgstr'):
                if not line:
                    raise Exception('malformed file at %d'% self.cur_line())
                source += unquote(line)
                line = self.lines.pop(0).strip()

            trad = unquote(line[7:])
            line = self.lines.pop(0).strip()
            while line:
                trad += unquote(line)
                line = self.lines.pop(0).strip()

            if targets and not fuzzy:
                # Use the first target for the current entry (returned at the
                # end of this next() call), and keep the others to generate
                # additional entries (returned the next next() calls).
                trans_type, name, res_id = targets.pop(0)
                code = trans_type == 'code'
                for t, n, r in targets:
                    if t == 'code' and code: continue
                    if t == 'code':
                        code = True
                    self.extra_lines.append((t, n, r, source, trad, comments))

        if name is None:
            if not fuzzy:
                _logger.warning('Missing "#:" formated comment at line %d for the following source:\n\t%s',
                                self.cur_line(), source[:30])
            return next(self)
        return trans_type, name, res_id, source, trad, '\n'.join(comments)
    __next__ = next

    def write_infos(self, modules):
        # import odoo.release as release
        self.buffer.write(u"# Translation of %(project)s.\n" \
                          "# This file contains the translation of the following modules:\n" \
                          "%(modules)s" \
                          "#\n" \
                          "msgid \"\"\n" \
                          "msgstr \"\"\n" \
                          '''"Project-Id-Version: %(project)s %(version)s\\n"\n''' \
                          '''"Report-Msgid-Bugs-To: \\n"\n''' \
                          '''"POT-Creation-Date: %(now)s\\n"\n'''        \
                          '''"PO-Revision-Date: %(now)s\\n"\n'''         \
                          '''"Last-Translator: <>\\n"\n''' \
                          '''"Language-Team: \\n"\n'''   \
                          '''"MIME-Version: 1.0\\n"\n''' \
                          '''"Content-Type: text/plain; charset=UTF-8\\n"\n'''   \
                          '''"Content-Transfer-Encoding: \\n"\n'''       \
                          '''"Plural-Forms: \\n"\n'''    \
                          "\n"

                          # % { 'project': release.description,
                          % { 'project': 'proj',
                              # 'version': release.version,
                              'version': '11',
                              'modules': ''.join("#\t* %s\n" % m for m in modules),
                              'now': datetime.utcnow().strftime('%Y-%m-%d %H:%M')+"+0000",
                            }
                          )

    def write(self, modules, tnrs, source, trad, comments=None):

        plurial = len(modules) > 1 and 's' or ''
        self.buffer.write(u"#. module%s: %s\n" % (plurial, ', '.join(modules)))

        if comments:
            self.buffer.write(u''.join(('#. %s\n' % c for c in comments)))

        code = False
        for typy, name, res_id in tnrs:
            self.buffer.write(u"#: %s:%s:%s\n" % (typy, name, res_id))
            if typy == 'code':
                code = True

        if code:
            # only strings in python code are python formated
            self.buffer.write(u"#, python-format\n")

        msg = (
            u"msgid %s\n"
            u"msgstr %s\n\n"
        ) % (
            quote(str(source)),
            quote(str(trad))
        )
        self.buffer.write(msg)





class DataGenerator:
    DEFAULT_USER_PASSWORD = '123'

    def __init__(self, env, default_user_password=DEFAULT_USER_PASSWORD):
        self.env = env
        self.default_user_password = default_user_password

    xml_ids = {
        'unit': 'uom.product_uom_unit',
        'main_company': 'base.main_company',
    }

    models = {
        'ConfigSettings': 'res.config.settings',
    }

    def account(self, company, code):
        return self.env['account.account'].search([
            ('company_id', '=', company.id),
            ('code', '=', code),
        ]).ensure_one()

    def execute_settings(self, vals):
        self.env['res.config.settings'].create(vals).execute()

    def create_property(self, model_name, field_name, res_id, company, value):
        field_id = self.env['ir.model.fields']._get_id(model_name, field_name)
        field_type = self.env[model_name]._fields[field_name].type

        # TODO Make res_id is list of ids
        property_vals = {
            'fields_id': field_id,
            'company_id': company.id,
            'res_id': res_id,
        }

        prop = self.env['ir.property'].search(
            [(name, '=', value) for name, value in property_vals.items()]
        )

        if prop:
            prop.write({
                'value': value,
                'type': field_type,
            })
        else:
            self.env['ir.property'].create(
                dict(property_vals, **{
                    'name': field_name,
                    'value': value,
                    'type': field_type,
                })
            )

    def create_user(self, vals, groups=()):

        self.env['res.users'].create(vals)

        # vals.setdefault('name', vals['login'].replace('_', ' ').replace('-', ' ').title())

        vals.setdefault('email', vals['login'] + '@example.com')
        vals.setdefault('password', self.default_user_password)

        vals.setdefault('company_ids', [(4, vals['company_id'])])

        vals.setdefault('country_id', self.russia.id)
        vals.setdefault('lang', 'ru_RU')
        vals.setdefault('tz', 'Europe/Saratov')

        if not vals.get('groups_id'):
            vals['groups_id'] = [self.env.ref(xml_id).id for xml_id in groups]

        return self.env['res.users'].with_context(no_reset_password=True).create(vals)

    # Main
    def generate(self):
        records = self.populate()
        self.save_refs(records)
        self.commit()

    def populate(self):
        pass

    def save_refs(self, records):
        if not records:
            return

        ModelData = env['ir.model.data']

        for name, value in records.items():
            if not isinstance(value, models.Model):
                continue

            if len(value) == 1 and not ModelData.search_count([
                ('model', '=', value._name),
                ('res_id', '=', value.id),
            ]):
                ModelData.create({
                    'module': 'test',
                    'name': name,
                    'model': value._name,
                    'res_id': value.id,
                })

    def commit(self):
        self.env.cr.commit()



import csv
import fnmatch
import os
import pathlib
from collections import defaultdict
from pprint import pprint

import odoo.api
# from odoo.tools import PoFile
from babel.messages import extract
from lxml import etree
from odoo.tools import xml_translate, html_translate

# from odoo_commands.config import read_config, odoo_project_config
# from .createdb import OdooProject
from odoo_commands.project import OdooProject, OdooModule
# from odoo_commands.odoo_translate import PoFile


# d = {
#     'menu': ('ir.ui.menu', ('name',)),
#     'act_window': ('ir.actions.act_window', ('name', 'help')),
#     'report': ('ir.actions.report', ('name', 'help')),
# }

class DataFileExtractor:
    tag_to_model = {
        'menuitem': 'ir.ui.menu',
        'act_window': 'ir.actions.act_window',
        'report': 'ir.actions.report',
    }

    translate_fields = {
        'ir.ui.view': {
            'name': False,
            'model': False,
            'key': False,
            'arch': xml_translate,
            'field_parent': False,
            'xml_id': False,
        },
        'ir.ui.menu': {
            'name': True,
            'web_icon': False,
        },
        'ir.actions.act_window': {
            'name': True,
            'type': False,
            'domain': False,
            'context': False,
            'res_model': False,
            'src_model': False,
            'view_mode': False,
            'usage': False,
            'search_view': False,
            'xml_id': False,
            'help': html_translate,
        },
        'ir.actions.report': {
            'name': True,
            'type': False,
            'model': False,
            'report_name': False,
            'report_file': False,
            'print_report_name': False,
            'attachment': False,
            'module': False,
            'lo_bin_path': False,
            'xml_id': False,
            'help': html_translate,
            # Added
            'py3o_template_fallback': False,
            'msg_py3o_report_not_available': False,
        },
        'ir.module.category': {
            'name': True,
            'description': True,
            'xml_id': False,
        },
        'res.groups': {
            'name': True,
            'comment': True,
        },
        'ir.model.access': {
            'name': False,
        },
        'ir.rule': {
            'name': False,
            'domain_force': False,
        },
    }

    # def __init__(self, field_translates: dict):
    #     self.field_translates = field_translates


    def extract_terms(self, module):
        with odoo.api.Environment.manage():
            registry = odoo.registry(f'module:{module.name}')

        self.res = self.extract_from_source_code(module)

    def extract_from_source_code(self, module: OdooModule):
        """
        Odoo 15: odoo.tools.translate.TranslationModuleReader._babel_extract_terms()
        """
        # result = []
        result = defaultdict(lambda: {
            'modules': set(),
            'tnrs': [],
            'comments': set(),
        })

        for method, path_template, keywords, options, extra_comments in [
            ('python', '**/*.py', {'_': None, '_lt': None}, {'encoding': 'UTF-8'}, []),
            # TODO Skip static/js/lib dir
            ('javascript', 'static/src/js/**/*.js', {'_t': None, '_lt': None}, None, ['openerp-web']),
            ('odoo.tools.translate:babel_extract_qweb', 'static/src/xml/**/*.xml', {'_': None}, None, ['openerp-web']),
            # ('odoo_commands.odoo_translate:babel_extract_qweb', 'static/src/xml/**/*.xml', {'_': None}, None, ['openerp-web']),
        ]:
            for file_path in module.path.glob(path_template):
                display_path = 'addons/' + str(file_path)
                with open(file_path, 'rb') as src_file:
                    for lineno, message, comments, _ in extract.extract(
                        method,
                        src_file,
                        keywords=keywords,
                        options=options,
                    ):
                        # result.append((module_name, trans_type, display_path, lineno, message, '', tuple(comments + extra_comments)))
                        message_data = result[message]
                        message_data['modules'].add(module.name)
                        message_data['tnrs'].append(('code', display_path, lineno))
                        message_data['comments'].update(comments + extra_comments)

        return result

    def extract_from_data_files(self, module):
        for data_file_path in module.data_file_paths():
            data_file_extension = data_file_path.suffix
            if data_file_extension == '.xml':
                self.extract_from_xml_data_file(data_file_path)
            elif data_file_extension == '.csv':
                self.extract_from_csv_data_file(data_file_path)

    def extract_from_xml_data_file(self, data_file_path):
        doc = etree.parse(data_file_path)
        root = doc.getroot()
        # return root

        def field_value(field_tag):
            if field_tag.get('type') == 'xml':
                return ''.join(etree.tostring(child, encoding='unicode') for child in field_tag)
            elif field_tag.text:
                return field_tag.text

        for record in root.iterfind('.//record'):
            model = record.get('model')
            for field in record.iterfind('./field'):
                field_name = field.get('name')
                value = field_value(field)
                if not value:
                    continue
                translate = self.translate(model, field_name)
                # TODO callable translate
                yield model, field.get('name'), value

            # if model in self.translate_fields:
            #     for field_name in self.translate_fields[model]:
            #         field_tag = record.find(f'./field[@name="{field_name}"]')
            #         if field_tag:
            #             value = field_value(field_tag)
            #             if value:
            #                 yield model, field_tag.get('name'), value
            # else:
            #     for field_tag in record.iterfind(f'./field'):
            #         value = field_value(field_tag)
            #         if value and translate:
            #             yield model, field_tag.get('name'), value

        for menuitem in root.iterfind('.//menuitem'):
            yield 'ir.ui.menu', 'name', menuitem.get('name')

        # TODO template tag
        # https://www.odoo.com/documentation/16.0/developer/reference/backend/data.html#template

    def extract_from_csv_data_file(self, file_path):
        model = file_path.stem
        with open(file_path) as csv_file:
            csv_reader = csv.reader(csv_file, quotechar='"', delimiter=',')
            headers = next(csv_reader)
            assert 'id' in headers

            translate_fields = set()
            for field_name in headers:
                if field_name.endswith(':id'):
                    continue
                translate = self.translate(model, field_name)
                if translate:
                    translate_fields.add(field_name)


def write_pot(modules, rows, pot_path, lang):
    buffer = open(pot_path, 'wb')
    writer = PoFile(buffer)
    writer.write_infos(modules)

    # we now group the translations by source. That means one translation per source.
    # grouped_rows = {}
    # for module, type, name, res_id, src, trad, comments in rows:
    #     row = grouped_rows.setdefault(src, {})
    #
    #     row.setdefault('modules', set()).add(module)
    #     if not row.get('translation') and trad != src:
    #         row['translation'] = trad
    #     row.setdefault('tnrs', []).append((type, name, res_id))
    #     row.setdefault('comments', set()).update(comments)

    for src, row in sorted(rows.items()):
        # if not lang:
            # translation template, so no translation value
            # row['translation'] = ''
        # elif not row.get('translation'):
        #     row['translation'] = src
        writer.write(row['modules'], row['tnrs'], src, '', row['comments'])


def parse_xml_file():
    doc = etree.parse(xmlfile)
    obj = xml_import(cr, module, idref, mode, report=report, noupdate=noupdate, xml_filename=xml_filename)
    obj.parse(doc.getroot(), mode=mode)


def generate_pot(module_paths, module_name):
    project = OdooProject(module_paths)
    module = project.module(module_name)

    res = extract_from_source_code(module)
    res.update(extract_from_data_files(module))

    for r in res:
        print(r, res[r])
    write_pot([module_name], res, 'test.pot', False)


# generate_pot(['tests/addons/'], 'module_name')



import contextlib
from functools import lru_cache


class Test:
    # def __init__(self):
    #     self._env = None

    # @cached_property
    @property
    @lru_cache()
    def env(self):
        def gen():
            print('1. before value')
            yield 'value'
            print('3. after value')

        self._env = gen()
        return next(self._env)

    def __del__(self):
        env_gen = getattr(self, '_env', None)
        if env_gen:
            next(env_gen, True)


# t = Test()
# print(t.env)
# print('2. payload')
# print(t.env)

@contextlib.contextmanager
def context():
    yield 'data'
    print('exit')


@contextlib.contextmanager
def fin_context():
    try:
        yield 'fin data'
    finally:
        print('exit')


class ClassContext:
    def __enter__(self):
        return 'class data'

    def __exit__(self, exc_type, exc_val, exc_tb):
        print('exit')


class TestWith:
    @property
    @lru_cache()
    def env(self):
        # with ClassContext() as data:
        #     return data

        def gen():
            with ClassContext() as data:
            # with context() as data:
            # with fin_context() as data:
                yield data

        # self.g = gen()
        # return next(self.g)
        # return next(gen())
        for g in gen():
            return g


    def payload(self):
        print(self.env)
        print('payload')


test = TestWith()
test.payload()
print('payload 2')



import functools
from functools import cached_property


# https://stackoverflow.com/questions/798442/what-is-the-correct-or-best-way-to-subclass-the-python-set-class-adding-a-new
def wrap(method):
    @functools.wraps(method)
    def wrapped_method(*args, **kwargs):
        return OdooModuleSet(method(*args, **kwargs))
    return wrapped_method


class OdooModuleSet(set):
    @cached_property
    def names_list(self):
        return sorted(module.name for module in self)

    @cached_property
    def names(self):
        return {module.name for module in self}

    def depends(self):
        return set().union(*[module.depends for module in self]) - self.names

    def sorted(self):
        return sorted(self, key=lambda m: m.name)

    for method in [
        '__or__',
        '__and__',
        '__sub__',
        'difference',
        'difference_update',
        'intersection_update',
        'symmetric_difference',
        'symmetric_difference_update',
        'intersection',
        'union',
        'copy',
    ]:
        exec(f'{method} = wrap(set.{method})')



import typer
from odoo_commands.createdb import create_database

app = typer.Typer()

app.command(name='createdb')(create_database)

@app.command()
def hello(name: str):
    print(f"Hello {name}")


if __name__ == "__main__":
    app()



import ast
from functools import cached_property, lru_cache
from pathlib import Path
import typing
from collections import ChainMap

from babel.core import Locale


def adapt_version(version, serie):
    """
    Copy-paste from odoo/modules/module.py:adapt_version()
    Adapt module version using Odoo major version.

    >>> adapt_version('1.0', '15.0')
    '15.0.1.0'
    >>> adapt_version('15.0', '15.0')
    '15.0.15.0'
    >>> adapt_version('15.0.1.0.0', '15.0')
    '15.0.1.0.0'
    """
    if version == serie or not version.startswith(serie + '.'):
        version = '%s.%s' % (serie, version)
    return version


def is_module(path: Path):
    return (path / '__manifest__.py').is_file()


class OdooModule:
    # TODO Add slots
    # __slots__ = ('path', 'name')

    README_FILE_NAMES = ['README.rst', 'README.md', 'README.txt']
    ICON_PATH = 'static/description/icon.png'

    # Default info from odoo/modules/module.py:load_information_from_description_file()
    default_info = {
        'application': False,
        'author': 'Odoo S.A.',
        'auto_install': False,
        'category': 'Uncategorized',
        'depends': [],
        'description': '',
        # 'icon': get_module_icon(module),
        'installable': True,
        'post_load': '',
        'version': '1.0',
        'web': False,
        'sequence': 100,
        'summary': '',
        'website': '',

        'data': [],
        'demo': [],
        'license': 'LGPL-3',
        # Ignored fields: test, init_xml, update_xml, demo_xml,
    }

    OTHER_FIELD_NAMES = frozenset([
        'name',
        'summary',
        'maintainer',
        'contributors',
        'external_dependencies',
        'to_buy',   # ?
        'assets',
        'pre_init_hook',
        'post_init_hook',
        'uninstall_hook',
        'images',
        'images_preview_theme',
        'snippet_lists',
        'live_test_url',
    ])

    _cache = {}

    def __new__(cls, path, *args, **kwargs):
        path = Path(path).resolve()
        if not is_module(path):
            raise ValueError()

        if path in cls._cache:
            return cls._cache[path]

        instance = super().__new__(cls, *args, **kwargs)
        instance.path = path
        instance.name = path.name
        cls._cache[path] = instance
        return instance

    def __repr__(self):
        return f'OdooModule({self.name!r})'

    def __eq__(self, other):
        return self.path == other.path

    def __hash__(self):
        return self.path.__hash__()

    def __truediv__(self, path):
        return self.path / path

    @cached_property
    def manifest(self):
        with open(self.path / '__manifest__.py') as manifest_file:
            return ast.literal_eval(manifest_file.read())

    def __getattr__(self, item):
        if item == 'name':
            raise ValueError('To get non-technical name of module use `shortdesc` attribute')
        elif item == 'shortdesc':
            item = 'name'

        if item in self.manifest:
            return self.manifest[item]
        elif item in self.default_info:
            return self.default_info[item]
        elif item in self.OTHER_FIELD_NAMES:
            return None

        raise AttributeError(f'Unknown attribute: {item!r}')

    @cached_property
    def description(self):
        module_description = self.__getattr__('description')
        if module_description:
            return module_description

        for readme_file_name in self.README_FILE_NAMES:
            readme_file_path = self / readme_file_name
            if readme_file_path.is_file():
                with open(readme_file_path) as description_file:
                    return description_file.read()

        return module_description

    @cached_property
    def icon(self):
        if 'icon' in self.manifest:
            return self.manifest['icon']
        # Copy-paste from odoo/modules/module.py:get_module_icon()
        if (self / self.ICON_PATH).is_file():
            return f'/{self.name}/{self.ICON_PATH}'
        return '/base/' + self.ICON_PATH

    @lru_cache
    def version(self, serie):
        return adapt_version(self.__getattr__('version'), serie)

    # TODO Read description from file

    def data_file_path(self):
        yield from self.manifest.get('data', [])

    # @property
    # def depends(self):
    #     return self.manifest.get('depends', [])

    @cached_property
    def auto_install(self):
        """Manifest `auto_install` can be True, False or list of depends. Convert it to bool"""
        auto_install = self.__getattr__('auto_install')
        if isinstance(auto_install, typing.Iterable):
            return True
        return auto_install

    @cached_property
    def auto_install_depends(self):
        auto_install = self.__getattr__('auto_install')
        if auto_install is False:
            return None
        if auto_install is True:
            return frozenset(self.depends)
        return frozenset(auto_install)

    def _file_translations(self, po_file_name):
        from odoo.tools.translate import PoFileReader

        result = {}
        for subdir in ['i18n_extra', 'i18n']:
            po_file_path = self / subdir / po_file_name
            if not po_file_path.exists():
                continue
            with open(po_file_path, 'rb') as f:
                for translation in PoFileReader(f):
                    trns = (
                        translation['type'],
                        translation['name'],
                        translation['res_id'],
                        translation['src'],
                    )
                    result[trns] = translation['value']

        return result

    @lru_cache()
    def translations(self, locale_identifier):
        locale = Locale.parse(locale_identifier)

        base_trans = self._file_translations(f'{locale.language}.po')
        if not locale.territory or locale.territory.lower() == locale.language:
            return base_trans

        # if locale.territory and locale.territory.lower() != locale.language:
        spec_trans = self._file_translations(f'{locale_identifier}.po')
        return ChainMap(spec_trans, base_trans)



import datetime

from odoo_commands.project import OdooModule, OdooModuleSet
from functools import lru_cache

SUPERUSER_ID = 1

class FakeDatabase:
    table_existing_query = """
        SELECT c.relname
          FROM pg_class c
          JOIN pg_namespace n ON (n.oid = c.relnamespace)
         WHERE c.relname IN %s
           AND c.relkind IN ('r', 'v', 'm')
           AND n.nspname = current_schema
    """

    read_module_fields_query = """SELECT "ir_module_module"."id" as "id", "ir_module_module"."name" as "name", "ir_module_module"."category_id" as "category_id", "ir_module_module"."shortdesc" as "shortdesc", "ir_module_module"."summary" as "summary", "ir_module_module"."description" as "description", "ir_module_module"."author" as "author", "ir_module_module"."maintainer" as "maintainer", "ir_module_module"."contributors" as "contributors", "ir_module_module"."website" as "website", "ir_module_module"."latest_version" as "latest_version", "ir_module_module"."published_version" as "published_version", "ir_module_module"."url" as "url", "ir_module_module"."sequence" as "sequence", "ir_module_module"."auto_install" as "auto_install", "ir_module_module"."state" as "state", "ir_module_module"."demo" as "demo", "ir_module_module"."license" as "license", "ir_module_module"."menus_by_module" as "menus_by_module", "ir_module_module"."reports_by_module" as "reports_by_module", "ir_module_module"."views_by_module" as "views_by_module", "ir_module_module"."application" as "application", "ir_module_module"."icon" as "icon", "ir_module_module"."to_buy" as "to_buy", "ir_module_module"."create_uid" as "create_uid", "ir_module_module"."create_date" as "create_date", "ir_module_module"."write_uid" as "write_uid", "ir_module_module"."write_date" as "write_date" FROM "ir_module_module" WHERE "ir_module_module".id IN %s"""

    module_table_fields = [
        'id',
        'name',
        'category_id',
        'shortdesc',
        'summary',
        'description',
        'author',
        'maintainer',
        'contributors',
        'website',
        'latest_version',
        'published_version',
        'url',
        'sequence',
        'auto_install',
        'state',
        'demo',
        'license',
        'menus_by_module',
        'reports_by_module',
        'views_by_module',
        'application',
        'icon',
        'to_buy',
        'create_uid',
        'create_date',
        'write_uid',
        'write_date',
    ]

    def __init__(self, installed_modules: OdooModuleSet):
        self.data = {}
        self.installed_modules = installed_modules
        self.fill_module_table(installed_modules)

    @staticmethod
    def _module_vals(module: OdooModule):
        return {
            'name': module.name,
            'shortdesc': module.shortdesc,
            'summary': module.summary,
            'description': module.description,
            'author': module.author,
            'maintainer': module.maintainer,
            'contributors': module.contributors,
            'website': module.website,
            'latest_version': module.version,
            'published_version': module.version,
            'url': module.live_test_url,
            'sequence': module.sequence,
            'auto_install': module.auto_install,
            'demo': None,
            'license': None,
            'menus_by_module': None,
            'reports_by_module': None,
            'views_by_module': None,
            'application': module.application,
            'icon': module.icon,
            'to_buy': module.to_buy,
        }

    def fill_module_table(self, installed_modules: OdooModuleSet):
        module_data = []
        now = datetime.datetime.now()
        for record_id, module in enumerate(installed_modules, start=1):
            module_vals = self._module_vals(module)
            module_vals.update({
                'id': record_id,
                'category_id': None,
                'state': 'installed',
                'create_uid': SUPERUSER_ID,
                'create_date': now,
                'write_uid': SUPERUSER_ID,
                'write_date': now,
            })
            module_data.append(module_vals)

        self.data['ir_module_module'] = module_data

    def execute(self, query, params) -> list[tuple | dict]:
        if query in {
            # self.table_existing_query,
            "SELECT proname FROM pg_proc WHERE proname='unaccent'",
            "SELECT * FROM ir_model_fields WHERE state='manual'",
            # Odoo 15
            "SELECT proname FROM pg_proc WHERE proname='word_similarity'",
            "SET SESSION lock_timeout = '15s'",
        }:
            return []

        if (query, params) in (
            ('SELECT * FROM ir_model WHERE state=%s', ['manual']),
        ):
            return []

        if query == self.table_existing_query and params == [('ir_module_module',)]:
            return [(1,)]

        if (
            query == "SELECT name from ir_module_module WHERE state IN %s"
            and params == (('installed', 'to upgrade', 'to remove'),)
        ):
            return [(module.name,) for module in self.installed_modules]

        # Odoo 15
        if query == 'SELECT "ir_module_module".id FROM "ir_module_module" WHERE ("ir_module_module"."state" = %s) ORDER BY  "ir_module_module"."name"  ':
            if params == ['installed']:
                return self.select('ir_module_module', ['id'], sort_key=lambda x: x['name'])

        if query == self.read_module_fields_query:
            return self.select('ir_module_module', self.module_table_fields, lambda x: x['id'] in params)

        if query == "SELECT name from ir_module_module WHERE state IN ('to install', 'to upgrade')":
            return []

        if query == "SELECT name, id, state, demo AS dbdemo, latest_version AS installed_version  FROM ir_module_module WHERE name IN %s":
            result = []
            for module in self.modules:
                if module['name'] in params:
                    module_copy = module.copy()
                    module_copy['dbdemo'] = module_copy.pop('demo')
                    result.append(module_copy)
            return result

        if query == 'select digits from decimal_precision where name=%s':
            return self.decimal_precision(params)

        # modules/loading.py:reset_modules_state
        if query in {
            "UPDATE ir_module_module SET state='installed' WHERE state IN ('to remove', 'to upgrade')",
            "UPDATE ir_module_module SET state='uninstalled' WHERE state='to install'",
        }:
            return []

        # ResLang._check_active
        if query == 'SELECT count(1) FROM "res_lang" WHERE ("res_lang"."active" = %s)':
            return [(1,)]

        # No test mode
        if query == "SELECT sequence_name FROM information_schema.sequences WHERE sequence_name='base_registry_signaling'":
            return [('base_registry_signaling',)]

        if query == """ SELECT base_registry_signaling.last_value,
                                  base_cache_signaling.last_value
                           FROM base_registry_signaling, base_cache_signaling""":
            return [(1, 1)]

        raise NotImplementedError(f'Unknown SQL query:\n{query}\n\nparams: {params}')

    def select(self, table, fields, condition=lambda x: True, sort_key=None, row_type=tuple):
        records = filter(condition, self.data[table])

        if sort_key:
            records = sorted(records, key=sort_key)

        result = []
        if row_type is tuple:
            for record in records:
                result.append(tuple(record[field] for field in fields))
        elif row_type is dict:
            for record in records:
                result.append({field: record[field] for field in fields})
        else:
            raise ValueError(f'Incorrect row_type: {row_type}')

        return result

    @property
    @lru_cache(maxsize=1)
    def modules(self):
        return [
            {
                'id': 1,
                'name': module.name,
                'state': 'installed',
                'dbdemo': False,
                'installed_version': '11.0.1.3',
            }
            for module_id, module in enumerate(self.installed_modules, start=1)
        ]

    def decimal_precision(self, names):
        return [(2,)] * len(names)


from odoo.sql_db import BaseCursor

class CursorMock(BaseCursor):
    """Mocked Odoo cursor class"""
    databases = {}

    def __init__(self, pool, dbname, dsn, serialized=True):
        self.dbname = dbname
        self.sql_log_count = 0
        self.transaction = None

    def execute(self, query, params=None, log_exceptions=None):
        database = self.databases.get(self.dbname)
        if not database:
            raise ValueError(f'There is no fake database: {self.dbname}')

        self.result = database.execute(query, params)

    @staticmethod
    def _convert_to_tuple(row):
        if isinstance(row, tuple):
            return row
        elif isinstance(row, dict):
            return tuple(row.values())

        raise TypeError(f'Unexpected type: {type(row)}')

    @staticmethod
    def _convert_to_dict(row):
        if isinstance(row, tuple):
            raise ValueError('Cursor execute result is list of tuples. Unavailable convert to dict')
        elif isinstance(row, dict):
            return row

        raise TypeError(f'Unexpected type: {type(row)}')

    def _fetchone(self, convert):
        if self.result:
            print(self.result)
            return convert(self.result.pop(0))

    def _fetchall(self, convert):
        fetch_result = list(map(convert, self.result))
        self.result = []
        return fetch_result

    def fetchone(self):
        return self._fetchone(self._convert_to_tuple)

    def fetchall(self):
        return self._fetchall(self._convert_to_tuple)

    def fetchmany(self, size=None):
        raise NotImplementedError

    def dictfetchone(self):
        self._fetchone(self._convert_to_dict)

    def dictfetchall(self):
        return self._fetchall(self._convert_to_dict)

    def dictfetchmany(self):
        raise NotImplementedError

    def commit(self):
        pass

    def close(self):
        pass

    def split_for_in_conditions(self, ids, size=None):
        return ids


def cursor_mock_class(installed_module_names_callback):

    class CursorDbMock(CursorMock):
        db = FakeDatabase(installed_module_names_callback)

    return CursorDbMock



from collections import OrderedDict
from pathlib import Path
from typing import List

from dataclasses import dataclass, field
import tomlkit

@dataclass(order=True)
class Config:
    project_module_dirs: List[str] = field(default_factory=list)
    third_party_module_dirs: List[str] = field(default_factory=list)
    include_modules: List[str] = field(default_factory=list)
    exclude_modules: List[str] = field(default_factory=list)

def read_config(path):
    with open(path) as config_file:
        config = tomlkit.load(config_file)

    return Config(**config.unwrap())



import time

from IPython import start_ipython

from odoo_commands.database_mock import CursorMock, FakeDatabase

s1 = time.time()

import logging
import mock

import odoo
from odoo.modules.registry import Registry
from odoo.tools import config

s2 = time.time()
print('01 >>', format(s2 - s1, '.2f'))


class MockRegistry(Registry):
    # def init(self, db_name):
    #     super().init(db_name)
    #     self._db = None

    def setup_signaling(self):
        print('setup_signaling')


class PostgresCursorMock:
    pass


class PostgresConnectionMock:
    def cursor(self):
        return PostgresCursorMock()


# class ConnectionPoolMock(ConnectionPool):
class ConnectionPoolMock:
    def __init__(self, maxconn=64):
        self._maxconn = max(maxconn, 1)

    def borrow(self, connection_info):
        return PostgresConnectionMock()


class ConnectionMock:
    pass

_logger = logging.getLogger(__name__)


def is_initialized(cr):
    return True

def reset_modules_state(dbname):
    print('reset_modules_state')

Registry.setup_signaling = lambda self: None
import odoo.modules
odoo.modules.reset_modules_state = lambda dbname: None


def start_env():
    t01 = time.time()
    config.parse_config([])
    # config.parse_config(['-d', 'prod2'])
    # config.parse_config(['--workers', '1'])
    odoo.cli.server.report_configuration()
    odoo.service.server.start(preload=[], stop=True)

    t02 = time.time()
    print('1 >>>', format(t02 - t01, '.2f'))

    class CursorDbMock(CursorMock):
        db = FakeDatabase(lambda self: ['base',])

    # with mock.patch('odoo.sql_db.db_connect', db_connect):
    with mock.patch('odoo.sql_db.Cursor', CursorDbMock):
        t03 = time.time()
        print('2 >>>', format(t03 - t02, '.2f'))

        # registry = odoo.registry('none')
        # registry = Registry('none')
        # cr = registry.cursor()
        # env = odoo.api.Environment(cr, odoo.SUPERUSER_ID, {})
        # registry = MockRegistry('none')

        with odoo.api.Environment.manage(), odoo.registry('none').cursor() as cr:
            env = odoo.api.Environment(cr, odoo.SUPERUSER_ID, {})
            t04 = time.time()
            print('3 >>>', format(t04 - t03, '.2f'))
            print('>>>>>', format(t04 - s1, '.2f'))
            start_ipython(argv=[], user_ns={'env': env})
            # yield env

        # with registry.cursor() as cr:
        #     env = odoo.api.Environment(cr, 1, {})

        # with odoo.api.Environment.manage():
            # registry = odoo.registry('none')
            # registry = MockRegistry('none')

s10 = time.time()
print('02 >>', format(s10 - s2, '.2f'))

start_env()



import os

from pathlib import Path
from functools import lru_cache, cached_property
from typing import Iterable, Container
import dataclasses

from odoo_commands.config import read_config, Config
from odoo_commands.module import OdooModule, is_module
from odoo_commands.module_set import OdooModuleSet


class OdooProject:
    def __init__(self, path='.', modules=None, **options):
        self.path = Path(path).resolve()
        self.modules = modules

        config_dict = dataclasses.asdict(self.project_file_config)
        config_dict.update(options)
        self.project_config = Config(**config_dict)

    @property
    def project_file_config(self):
        return read_config(self.path / 'odoo-project.toml')

    @property
    def project_module_paths(self):
        return [(self.path / path) for path in self.project_config.project_module_dirs]

    @property
    def third_party_module_paths(self):
        return [(self.path / path) for path in self.project_config.third_party_module_dirs]

    @property
    @lru_cache()
    def modules_paths(self):
        # In fact core_modules_path is os.path.join(odoo.tools.config['root_path'], 'addons').
        # We don't want to import odoo. It's long one second almost.
        import importlib
        odoo_path = importlib.util.find_spec('odoo').submodule_search_locations[0]
        core_modules_path = Path(odoo_path, 'addons').resolve()

        return (
            self.project_module_paths
            + self.third_party_module_paths
            + [core_modules_path]
        )

    def collect_modules(self, paths: Iterable[Path], exclude: Container[str] = ()):
        modules = OdooModuleSet()
        for module_path in paths:
            for subdir in module_path.iterdir():
                if subdir.name not in exclude and is_module(subdir):
                    modules.add(OdooModule(subdir))
        return modules

    @lru_cache(maxsize=1024)
    def module(self, module_name):
        for modules_path in self.modules_paths:
            module_path = modules_path / module_name
            if is_module(module_path):
                # TODO Cache?
                return OdooModule(module_path)
        raise LookupError(f'No module found: {module_name}')

    def find_modules(self, module_names):
        return OdooModuleSet(self.module(module_name) for module_name in module_names)

    @property
    def required_modules(self):
        if self.modules:
            return self.find_modules(self.modules)
        return (
            self.collect_modules(self.project_module_paths, exclude=self.project_config.exclude_modules)
            | self.find_modules(self.project_config.include_modules)
        )

    def expand_dependencies(self, modules):
        modules_to_check_depends = modules

        while True:
            module_depends_names = modules_to_check_depends.depends()
            if module_depends_names:
                modules_to_check_depends = self.find_modules(module_depends_names)
                modules |= modules_to_check_depends
            else:
                break

        return modules

    def auto_install_modules(self, modules: OdooModuleSet):
        auto_install_modules = OdooModuleSet()
        for module in self.collect_modules(self.modules_paths, exclude=modules.names):
            if module.auto_install and set(module.depends) <= modules.names:
                auto_install_modules.add(module)
        return auto_install_modules

    @property
    def installed_modules(self):
        modules = self.expand_dependencies(self.required_modules)

        while True:
            auto_install_modules = self.auto_install_modules(modules)
            if auto_install_modules:
                modules |= auto_install_modules
            else:
                break

        return modules

    def topologic_dependencies(self, modules: OdooModuleSet):
        result = []
        visited = set()

        def visit(modd):
            if modd not in visited:
                visited.add(modd)
                if modd in modules:
                    for depend_module in self.find_modules(modd.depends).sorted():
                        visit(depend_module)
                    result.append(modd)

        for module in modules.sorted():
            visit(module)

        return result

    @lru_cache()
    def installed_modules_list(self, inversed=False):
        modules = self.topologic_dependencies(self.installed_modules)
        if inversed:
            modules = modules.reverse()
        return modules

    # =======================================================================

    # def field(self, module_name, model_name, field_name):
    #     module = self.module(module_name)
    #     field = Field()
    #     # for module in module.topological_depends():
    #     for module in self.topological_depends(module):
    #         for model_def in module.model_defs(model_name):
    #             for attr_def in model_def.attr_defs(field_name):
    #                 field.apply(attr_def)
    #     return field

    def topological_depends(self, module):
        # expanded_modules = self.expanded(module)
        graph = self.module_graph(module)
        return topologic(graph)

    def module_graph_OFF(self, module_name, processed_modules=set()):
        edges = []
        processed_modules.add(module_name)
        module = self.module(module_name)
        edges.extend([(module, depend_module) for depend_module in module.depends])
        for dependency_module_name in module.depends:
            if dependency_module_name not in processed_modules:
                edges.extend(self.module_graph(dependency_module_name, processed_modules))
        return edges

    def module_graph(self, module_name):
        processed_modules = set()

        def recursive_module_graph(module):
            edges = []
            processed_modules.add(module)
            for dependency_module_name in module.depends:
                dependency_module = self.module(dependency_module_name)
                edges.append((module, dependency_module))
                if dependency_module not in processed_modules:
                    edges.extend(self.module_graph(dependency_module))
            return edges

        return recursive_module_graph(self.module(module_name))

    # @lru_cache(maxsize=1024)
    # def module(self, module_name):
    #     return Module(self.module_path(module_name), module_name)

    def attr(self, model_name, attr_name, module=None):
        deps = self.module_dependencies(module_name)

    # ============= create db ====================
    # Copy of click_odoo_contrib/initdb.py:addons_hash
    def addons_hash(self, module_names, with_demo):
        h = hashlib.sha1()
        h.update("!demo={}!".format(int(bool(with_demo))).encode("utf8"))
        for module_name in sorted(module_names):
            module_path = self.module_path(module_name)
            h.update(module_name.encode("utf8"))
            for filepath in _walk(module_path):
                h.update(filepath.encode("utf8"))
                with open(os.path.join(module_path, filepath), "rb") as f:
                    h.update(f.read())
        return h.hexdigest()

    @lru_cache(maxsize=1024)
    def module_mtime(self, module_name):
        module_path = self.module_path(module_name)
        return max(os.path.getmtime(file_path) for file_path in _walk(module_path))

    def read_dependencies(self, module_names: set):
        for module_name in module_names:
            pass

    def cache_key(self, module_names, timestamp):
        module_mtime = {}
        for module_name in module_names:
            module_path = self.path(module_name)


    # Env
    # TODO Use functools.cached_property for Python3.8+
    @property
    @lru_cache(maxsize=1)
    def env(self):
        # Keep ref on generator force close generator later
        # self._env_generator = self._get_env_generator()
        # return next(self._env_generator)
        return self._get_env('soma', self.installed_modules)

    def _get_env(self, db_name, installed_modules: OdooModule | OdooModuleSet):
        import mock
        from odoo_commands.database_mock import CursorMock, FakeDatabase
        import odoo

        CursorMock.databases[db_name] = FakeDatabase(installed_modules)

        with mock.patch('odoo.sql_db.Cursor', CursorMock):
            with odoo.registry(db_name).cursor() as cr:
                return odoo.api.Environment(cr, odoo.SUPERUSER_ID, {})

    # ========== Translation ==========
    def search_translation(self, tupl, lang=None):
        for module in self.installed_modules_list():
            translations = module.translations(lang or self.lang)
            if tupl in translations:
                return translations[tupl]

    def code_translation(self, file_path, value, lineno=None):
        # if not file_path.startswith('addons/'):
        #     file_path = f'addons/{file_path}'
        return self.search_translation(('code', file_path, lineno, value))

    def selection(self, model_name, field_name, value, lang=None):
        field = self.env[model_name]._fields[field_name]
        for selection_value, selection_string in field.selection:
            if value == selection_value:
                break
        else:
            raise ValueError('Invalid selection value')
        return self.search_translation(('selection', f'{model_name},{field_name}', '0', selection_string), lang=lang)

    def constraint(self, model, value):
        return self.search_translation(('constraint', model, '0', value))

    def sql_constraint(self, model, value):
        return self.search_translation(('sql_constraint', model, '0', value))

    def model_translation(self, model_name, field_name, xml_id, origin, lang=None):
        return self.search_translation(('model', f'{model_name},{field_name}', xml_id, origin), lang=lang)

    def model_description(self, model_name, lang=None):
        model = self.env[model_name]
        if not lang:
            return model._description
        xml_id = f'{model._module}.model_{model_name.replace(".", "_")}'
        return self.model_translation('ir.model', 'name', xml_id, model._description, lang=lang)

    # def field_description(self, module_name, model_name, field_name):
    def field_description(self, model_name, field_name, lang=None):
        field = self.env[model_name]._fields[field_name]
        if not lang:
            return field.string
        xml_id = f'{field._module}.field_{model_name.replace(".", "_")}_{field_name}'
        return self.model_translation('ir.model.fields', 'field_description', xml_id, field.string, lang=lang)



import ast
import glob
import hashlib
import os
import datetime
import runpy
import subprocess
from functools import lru_cache
from pprint import pprint
from typing import Set
import time

# import odoo

# from click_odoo_contrib.initdb import _walk

import logging

from odoo_commands.module_set import OdooModuleSet
from odoo_commands.project import OdooProject

logger = logging.getLogger(__name__)


class IndentLogger(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        indent_level = self.extra['indent_level']
        return '    ' * indent_level + msg, kwargs

SECONDS_IN_HOUR = 60 * 60
SECONDS_IN_DAY = 24 * 60 * 60
EPOCH_FIRST_MONDAY = SECONDS_IN_DAY * 4

# import odoo
def shift(t, delta):
    # return (t + EPOCH_FIRST_MONDAY) - t % delta - EPOCH_FIRST_MONDAY
    # return t - (t + EPOCH_FIRST_MONDAY) % delta - EPOCH_FIRST_MONDAY
    return t - (t + EPOCH_FIRST_MONDAY) % delta
    # return (t - EPOCH_FIRST_MONDAY) - t % delta + EPOCH_FIRST_MONDAY


def cache_time_point_generator_3(seq):
    # SECONDS_IN_HOUR = 60 * 60
    # SECONDS_IN_DAY = 24 * 60 * 60
    # EPOCH_FIRST_MONDAY = SECONDS_IN_DAY * 4

    now = int(time.time())
    now = int(time.time() - 60 * SECONDS_IN_DAY)
    # now = 0
    # t = now + EPOCH_FIRST_MONDAY

    seq = [
        SECONDS_IN_HOUR,
        12 * SECONDS_IN_HOUR,
        SECONDS_IN_DAY,
        7 * SECONDS_IN_DAY,
        4 * 7 * SECONDS_IN_DAY,
        24 * 7 * SECONDS_IN_DAY,
    ]

    prev = None
    for s in seq:
        current = shift(now, s)
        if current != prev:
            yield datetime.datetime.fromtimestamp(current)
            prev = current

    # last_hour = shift(now, SECONDS_IN_HOUR)
    # last_12_hours = shift(now, 12 * SECONDS_IN_HOUR)
    # # last_hour = t - t % SECONDS_IN_HOUR
    # # last_12_hours = t - t % (12 * SECONDS_IN_HOUR)
    # last_day = shift(now, SECONDS_IN_DAY)
    # last_sunday = shift(now, 7 * SECONDS_IN_DAY)
    # last_4th_sunday = shift(now, 4 * 7 * SECONDS_IN_DAY)
    # last_24th_sunday = shift(now, 24 * 7 * SECONDS_IN_DAY)
    #
    # return [
    #     datetime.datetime.fromtimestamp(last_hour),
    #     datetime.datetime.fromtimestamp(last_12_hours),
    #     datetime.datetime.fromtimestamp(last_day),
    #     datetime.datetime.fromtimestamp(last_sunday),
    #     datetime.datetime.fromtimestamp(last_4th_sunday),
    #     datetime.datetime.fromtimestamp(last_24th_sunday),
    # ]

def cache_time_point_generator_2():
    now = datetime.datetime.utcnow()
    today = now.date()
    # weekday is in [0,6]; 0 is Monday
    last_monday = today - datetime.timedelta(days=today.weekday())

    first_monday = datetime.date(1970, 1, 5)
    days_from_first_monday = (last_monday - first_monday).days

    last_4th_sunday = last_monday - datetime.timedelta(days=days_from_first_monday % (7 * 4))
    last_24th_sunday = last_monday - datetime.timedelta(days=days_from_first_monday % (7 * 24))

    year, month, day, hour, *_ = now.timetuple()
    return [
        datetime.datetime(year, month, day, hour, 0, 0, 0),
        datetime.datetime(year, month, day, 0, 0, 0, 0),
        last_monday,
        last_4th_sunday,
        last_24th_sunday,
    ]


def cache_time_point_generator(dt=None):
    if dt:
        now = dt
    else:
        # now = datetime.datetime.now()
        now = datetime.datetime.utcnow()

    today = now.date()
    # weekday is in [0,6]; 0 is Monday
    last_monday = today - datetime.timedelta(days=today.weekday())

    first_monday = datetime.date(1970, 1, 5)
    days_from_first_monday = (last_monday - first_monday).days

    last_4th_sunday = last_monday - datetime.timedelta(days=days_from_first_monday % (7 * 4))
    last_24th_sunday = last_monday - datetime.timedelta(days=days_from_first_monday % (7 * 24))

    year, month, day, hour, *_ = now.timetuple()
    return [
        datetime.datetime(year, month, day, hour, 0, 0, 0),
        datetime.datetime(year, month, day, 0, 0, 0, 0),
        last_monday,
        last_4th_sunday,
        last_24th_sunday,
    ]

    return [
        time.mktime((year, month, day, hour, 0, 0, 0, 0, 0)),
        time.mktime((year, month, day, 0, 0, 0, 0, 0, 0)),
        time.mktime(last_monday.timetuple()),
        time.mktime(last_4th_sunday.timetuple()),
        time.mktime(last_24th_sunday.timetuple()),
    ]

# pprint(list(cache_time_point_generator_3([])))


# def read_manifest(module_dir):
#     manifest_path = os.path.join(module_dir, '__manifest__.py')
#     # if not os.path.isfile(manifest_path):
#     #     raise FileNotFoundError("No Odoo manifest found in %s" % addon_dir)
#     with open(manifest_path) as manifest_file:
#         return ast.literal_eval(manifest_file.read())


contrib_module_path = '/home/voronin/.local/share/virtualenvs/sintez_addons-7QRHjYmJ/lib/python3.6/site-packages/odoo/addons'

def module_dependencies(module_dir):
    return read_manifest(module_dir).get('depends', [])


def contrib_module_deps(contrib_module_path):
    res = {}
    for module_dir in glob.iglob(contrib_module_path + '/*'):
        if module_dir.endswith('__pycache__') or module_dir.endswith('__init__.py'):
            continue
        res[module_dir] = module_dependencies(module_dir)
    return res


def get_cache_timestamp_modules(modules, cache_timestamp):
    pass


def install(modules, cache_timestamps, level=0):
    # indent_logger = IndentLogger(logger, {'indent_level': level})
    cache_timestamp = next(cache_timestamps)
    cache_timestamp_modules = get_cache_timestamp_modules(modules, cache_timestamp)
    h = hash(cache_timestamp_modules)
    database = get_database(h)
    if not database:
        database = install(cache_timestamp_modules, cache_timestamps)
    return install_modules(database, modules - cache_timestamp_modules)


def install_modules(database, modules_to_install: OdooModuleSet):
    new_database = name()
    # cache
    copy_database(database, new_database)
    install_modules_2(new_database, modules_to_install)
    return new_database


def create_database(database: str, populate=True):
    project = OdooProject()
    modules_to_install = project.expand_dependencies(project.required_modules)
    modules = ','.join(modules_to_install.names_list)

    subprocess.check_call([
        'odoo',
        '--no-http',
        '--stop-after-init',
        f'--database={database}',
        f'--init={modules}',
        '--without-demo=all',
    ])

    if populate:
        import runpy
        global_vars = runpy.run_path(project.path / 'tests/populate_database.py')
        populate_database = global_vars['populate_database']

        import odoo

        registry = odoo.registry(database)
        with registry.cursor() as cr:
            env = odoo.api.Environment(cr, odoo.SUPERUSER_ID, {})
            populate_database(env)
            env.cr.commit()


    # cache_time_points = cache_time_point_generator()
    # return install(project.expand_dependencies(project.required_modules), cache_time_points)



from lxml import etree


def convert_xml_import(xml_file_path):
    doc = etree.parse(xml_file_path)
    root = doc.getroot()
    # return root

    for record in root.iterfind('.//record'):
        model = record.get('model')
        for field in record.iterfind('./field'):
            print(field.get('name', 'No'))
            # value = None
            # if field.get('type') == 'xml':
            #     value = '\n'.join(etree.tostring(element) for element in field)
            # elif field.text:
            #     value = field.text
            #
            # if value:
            #     yield model, field.get('name'), value

    # d = {
    #     'menu': ('ir.ui.menu', ('name',)),
    #     'act_window': ('ir.actions.act_window', ('name', 'help')),
    #     'report': ('ir.actions.report', ('name', 'help')),
    # }

    # for tag, model, fields in d:
    #     element = root.iterfind(tag)
    #     for field in fields:
    #         yield model, field, element.get(field)


root = convert_xml_import('tests/addons/module_name/views/sale_order.xml')
